/*
 * 
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA.
 * 
 * 
 */
 


#include <iostream>
#include <random>
#include <chrono>
#include <set>
#include <map>
#include <list>

#include "Matrix.h"

using namespace std;

#define PRINT_INFO (true)

unsigned seed = std::chrono::system_clock::now().time_since_epoch().count();
default_random_engine rand_generator (seed);

Matrix<uint8_t> GenRandomMat(unsigned int n)
{
	Matrix<uint8_t> mat1 (n,n,0);
	Matrix<uint8_t> mat2 (n,n,0);
	Matrix<uint8_t> res (n,n,0);
	

	uniform_int_distribution<int> distribution(0,1);
	
	for (unsigned int i = 0; i < n; i++)
	{
		for (unsigned int j = i+1; j < n; j++) mat1(i,j) = distribution(rand_generator);
		for (unsigned int j = 0; j < i; j++) mat2(i,j) = distribution(rand_generator);
		//for (unsigned int j = i+1; j < n; j++) mat1(i,j) = 0;
		//for (unsigned int j = 0; j < i; j++) mat2(i,j) = 0;
		mat1(i,i) = 1;
		mat2(i,i) = 1;
	}
	
	for (unsigned int i = 0; i < n; i++)
	{
		for (unsigned int j = 0; j < n; j++)
		{
			for (unsigned int k = 0; k < n; k++) res(i,j) ^= mat1(i,k) & mat2(k,j);
		}
	}
	return res;
}

class Poly;

class Term
{
	public:
		Term() {};
		Term(uint8_t var) {m_set.emplace(var);};
		Term(set<uint8_t> const & s) : m_set(s) {};
		Term(set<uint8_t> && s) : m_set(move(s)) {};
		template <typename iter> Term(iter it, iter end) : m_set(it,end) {};
		Term(Term const & t) : m_set(t.m_set) {};
		Term(Term && t) : m_set(move(t.m_set)) {};
		
		Term & operator=(Term const & t) {m_set = t.m_set; return *this;};
		Term & operator=(Term && t) {m_set = move(t.m_set); return *this;};
		
		~Term() {};
		
		bool isCst() const {return m_set.empty();};
		
		set<uint8_t> const & getVars() const {return m_set;};
		
		bool replace(uint8_t old, uint8_t var) {
			auto it = m_set.find(old);
			if (it != m_set.end())
			{
				m_set.erase(it);
				m_set.emplace(var);
				return true;
			}
			else return false;
		};
		
		bool set0Var(uint8_t var) {
			auto it = m_set.find(var);
			if (it != m_set.end())
			{
				m_set.clear();
				return true;
			}
			else return false;
		};
		
		bool set1Var(uint8_t var) {
			auto it = m_set.find(var);
			if (it != m_set.end())
			{
				m_set.erase(it);
				return true;
			}
			else return false;
		};
		
		
		
		unsigned int degree() const {return m_set.size();};
		
		Term & operator*=(Term const & t) {m_set.insert(t.m_set.begin(),t.m_set.end()); return *this;};
		
		friend bool operator==(Term const & t1, Term const & t2) {
			if (t1.m_set.size() != t2.m_set.size()) return false;
			auto it1 = t1.m_set.begin();
			auto const end1 = t1.m_set.end();
			auto it2 = t2.m_set.begin();
			while (it1 != end1)
			{
				if (*it1 != *it2) return false;
				++it1;
				++it2;
			}
			return true;
		};
		
		friend bool operator!=(Term const & t1, Term const & t2) {return !(t1 == t2);};
		friend bool operator<(Term const & t1, Term const & t2) {
			auto const s1 = t1.m_set.size();
			auto const s2 = t2.m_set.size();
			if (s1 == s2)
			{
				auto it1 = t1.m_set.begin();
				auto const end1 = t1.m_set.end();
				auto it2 = t2.m_set.begin();
				while (it1 != end1)
				{
					if (*it1 == *it2)
					{
						++it1;
						++it2;
					}
					else return (*it1 < *it2);
				}
				return false;
			}
			else return (s1 < s2);
		};
		
		friend Term operator*(Term const & t1, Term const & t2) {auto res = t1; res *= t2; return res;};
		
		void print() const {
			if (this->isCst()) cout << 1;
			else
			{
				auto it = m_set.begin();
				cout << "x" << static_cast<unsigned int>(*it);
				++it;
				while (it != m_set.end())
				{
					cout << "*x" << static_cast<unsigned int>(*it);
					++it;
				}
			}
		};
		
		friend Poly operator+(Term const & t1, Term const & t2);
		
	private:
		set<uint8_t> m_set;
		
		friend Poly;
};

class Poly
{
	public:
		Poly() {};
		Poly(Term const & t) {m_set.emplace(t);};
		Poly(Term && t) {m_set.emplace(move(t));};
		Poly(Poly const & p) : m_set(p.m_set) {};
		Poly(Poly && p) : m_set(move(p.m_set)) {};
		
		Poly & operator=(Poly const & p) {m_set = p.m_set; return *this;};
		Poly & operator=(Poly && p) {m_set = move(p.m_set); return *this;};
		
		template <typename iter> Poly(iter it, iter end) : m_set(it,end) {};
		~Poly() {};
		
		set<Term> const & getTerms() const {return m_set;};
		set<Term> & getTerms() {return m_set;};
		
		Term const & highestTerm() const {return *(m_set.rbegin());};
		bool isCst() const {
			if (m_set.empty()) return true;
			else
			{
				if (m_set.size() > 1) return false;
				else return (m_set.begin()->isCst());
			}
		};
		
		bool isNull() const {return m_set.empty();};
		
		Poly & operator+=(Poly const & p) {
			for (auto const & t : p.m_set) *this += t;
			return *this;
		};
		
		Poly & operator+=(Term const & t) {
			if (m_set.count(t) == 1) m_set.erase(t);
			else m_set.insert(t);
			return *this;
		};
		
		Poly & operator+=(uint8_t cst) {
			if (cst == 1) *this += Term();
			return *this;
		};
		
		Poly & operator*=(Poly const & p) {
			Poly tmp;
			for (auto const & t : m_set) tmp += p*t;
			m_set = move(tmp.m_set);
			return *this;
		};
		Poly & operator*=(Term const & t) {
			Poly tmp;
			for (auto const & tt : m_set) tmp += tt*t;
			m_set = move(tmp.m_set);
			return *this;
		};
		
		uint8_t evaluate(vector<uint8_t> const & vec) const {
			uint8_t res = 0;
			for (auto const & t : m_set)
			{
				auto const & vars = t.getVars();
				auto const end = vars.end();
				auto it = vars.begin();
				while (it != end && vec[*it] != 0)
				{
					if (vec[*it] == 2) cout << "strange";
					++it;
				}
				if (it == end) res ^= 1;
			}
			return res;
		};
		
		friend bool operator==(Poly const & p1, Poly const & p2) {return (p1.m_set == p2.m_set);};
		
		friend bool operator!=(Poly const & p1, Poly const & p2) {return !(p1 == p2);};
		
		friend Poly operator+(Poly const & p, Term const & t) {auto res = p; res+=t; return res;};
		friend Poly operator+(Term const & t1, Term const & t2) {Poly res = t1; res+=t2; return res;};
		friend Poly operator+(Poly const & p1, Poly const & p2) {auto res = p1; res+=p2; return res;};
		
		friend Poly operator*(Poly const & p, Term const & t) {auto res = p; res*=t; return res;};
		friend Poly operator*(Poly const & p1, Poly const & p2) {auto res = p1; res*=p2; return res;};
		
		void print() const {
			if (m_set.empty()) cout << 0;
			else
			{
				for (auto const & t : m_set)
				{
					t.print();
					cout << " + ";
				}
			}
		};
		
	private:
		set<Term> m_set;
		
};

Poly Replace(uint8_t old, Poly const & G, Poly const & P)
{	
	Poly res;
	auto const & set_terms = P.getTerms();
	for (auto const & t : set_terms)
	{
		auto const & set_vars = t.getVars();
		if (set_vars.count(old) == 0) res += t;
		else
		{
			auto tmp = set_vars;
			tmp.erase(old);
			res += G*Term(move(tmp));
		}
	}
	return res;
}

Poly ReplaceNeg(uint8_t old, uint8_t var, Poly const & P)
{
	Poly res;
	auto const & set_terms = P.getTerms();
	for (auto const & t : set_terms)
	{
		Term tt = t;
		if (tt.set1Var(old))
		{
			res += tt;
			res += tt*Term(var);
		}
		else res += t;
	}
	return res;
}

Poly Set0Var(uint8_t var, Poly const & P)
{
	Poly res;
	auto const & set_terms = P.getTerms();
	for (auto const & t : set_terms)
	{
		auto const & set_vars = t.getVars();
		if (set_vars.count(var) == 0) res += t;
	}
	return res;
}

Poly Set1Var(uint8_t var, Poly const & P)
{	
	Poly res;
	auto const & set_terms = P.getTerms();
	for (auto const & t : set_terms)
	{
		Term tt = t;
		if (tt.set1Var(var))
		{
			auto const & set_vars = tt.getVars();
			if (set_vars.empty()) res += 1;
			else res += tt;
		}
		else res += tt;
	}
	return res;
}

vector<Poly> AffineLayer(Matrix<uint8_t> const & mat, vector<Poly> const & in)
{
	uniform_int_distribution<int> distribution(0,1);
	
	const unsigned int n = mat.getNbLines();
	vector<Poly> res (n);
	
	for (unsigned int i = 0; i < n; i++)
	{
		for (unsigned int j = 0; j < n; j++)
		{
			if (mat(i,j) == 1) res[i] += in[j];
		}
		res[i] += distribution(rand_generator);
	}
	
	return res;
}

vector<Poly> NonLinLayer(vector<Poly> const & in)
{
	const unsigned int n = in.size();
	vector<Poly> res (n);
	
	for (unsigned int i = 0; i < n; i++)
	{
		res[i] = in[i] + in[(i+2)%n] + (in[(i+1)%n]*in[(i+2)%n]);
	}
	
	return res;
}

bool simplifySystem(list<Poly> & my_system)
{
	auto const end = my_system.end();
	auto it = my_system.begin();
	while (it != end)
	{
		if (it->isCst())
		{
			if (it->isNull()) it = my_system.erase(it);
			else return false;
		}
		else
		{
			auto const & t = it->highestTerm();
			auto it2 = my_system.begin();
			while (it2 != end)
			{
				if (it2 != it)
				{
					auto & p = *it2;
					auto const & tmp_set = p.getTerms();
					if (tmp_set.count(t) != 0)
					{
						p += *it;
						if (p.isCst())
						{
							if (p.isNull()) it2 = my_system.erase(it2);
							else return false;
						}
						else ++it2;
					}
					else ++it2;
				}
				else ++it2;
			}
			++it;
		}
	}
	return true;
}


vector<Poly> GeneratePublicKey(unsigned int n, unsigned int nb_perturbations)
{
	auto mat1 = GenRandomMat(n);
	auto mat2 = GenRandomMat(n);
	auto mat3 = GenRandomMat(n);
	
	vector<Poly> my_vec;
	for (unsigned int i = 0; i < n; i++) my_vec.emplace_back(i);
	
	vector<Poly> tmp1 = AffineLayer(mat1,my_vec);
	my_vec.clear();
	vector<Poly> tmp2 = NonLinLayer(tmp1);
	tmp1.clear();
	vector<Poly> tmp3 = AffineLayer(mat2,tmp2);
	tmp2.clear();
	
	cout << "First linear layer: " << endl;
	for (unsigned int i = 0; i < n; i++)
	{
		for (unsigned int j = 0; j < n; j++) cout << static_cast<unsigned int>(mat1(i,j));
		cout << endl;
	}
	cout << endl;
	
	return tmp3;
}


list<list<unsigned int>> FindFirstLayerCHAINS(const unsigned int pos0, const unsigned int pos1, const unsigned int old, const unsigned int n, map<unsigned int, set<unsigned int>> const & map_cycles, map<unsigned int, unsigned int> & map_valid)
{
	list<list<unsigned int>> res;
	auto const & s0 = map_cycles.at(pos0);
	auto const & s1 = map_cycles.at(pos1);
	for (auto x : s0)
	{
		auto & valid = map_valid[x];
		if (valid <= 1)
		{
			if (valid == 1 && old == n+1) 
			{
				list<unsigned int> L;
				L.emplace_back(pos0);
				res.emplace_back(move(L));
			}
			if (valid == 0 && old <= n && s1.count(x) != 0)
			{
				valid = old;
				auto res_tmp = FindFirstLayerCHAINS(x,pos0,old+1,n,map_cycles,map_valid);
				for (auto & L : res_tmp) L.emplace_back(pos0);
				res.splice(res.begin(),res_tmp);
				valid = 0;
			}
		}
	}
	return res;
}

list<vector<uint8_t>> FindFirstLayerLINSOLVER(list<Poly> & my_system, const unsigned int n)
{
	list<vector<uint8_t>> res;
	
	{
		list<pair<uint8_t,Poly>> linear;
		
		auto it = my_system.begin();
		auto const end = my_system.end();
		while (it != end)
		{
			auto const & t = it->highestTerm();
			if (t.degree() == 1)
			{
				linear.emplace_back(make_pair(*(t.getVars().begin()),t + *it));
				it = my_system.erase(it);
			}
			else ++it;
		}
		
		{
			res.emplace_back(vector<uint8_t>(n,2));
			auto const end_lin = linear.end();
			for (auto it_lin = linear.begin(); it_lin != end_lin ; ++it_lin)
			{
				auto const & terms = it_lin->second.getTerms();
				set<uint8_t> vars;
				for (auto const & t : terms)
				{
					if (t.degree() == 1) vars.emplace(*(t.getVars().begin()));
				}
				auto it_res = res.begin();
				while (it_res != res.end())
				{
					auto & my_vec = *it_res;
					for (auto const & x : vars)
					{
						if (my_vec[x] == 2)
						{
							my_vec[x] = 1;
							res.emplace_back(my_vec);
							my_vec[x] = 0;
						}
					}
					my_vec[it_lin->first] = it_lin->second.evaluate(my_vec);
					++it_res;
				}
			}
			res.erase(res.begin());
		}
		
	}
	return res;
}

list<vector<uint8_t>> FindFirstLayerSTEP1(vector<map<unsigned int, Poly>> & my_maps, const unsigned int n, unsigned int pos0, unsigned int pos1, unsigned int pos2, list<Poly> & my_system)
{
	static long unsigned int cpt = 0;
	cout << "\r" << "complexity: " << ++cpt << " operations ";
	if (simplifySystem(my_system) == false) return list<vector<uint8_t>> ();
	auto const nb_eq = my_system.size();
	if (nb_eq == n) return list<vector<uint8_t>> ();
	if (pos0 >= my_maps.size()) return FindFirstLayerLINSOLVER(my_system,n);
	if (nb_eq == n - 1)
	{
		auto res = FindFirstLayerLINSOLVER(my_system,n);
		auto const & v = *(res.begin());
		for (; pos0 < my_maps.size(); ++pos0)
		{
			if (pos1 == ~0u)
			{
				unsigned int k = 0;
				while (k < n && my_maps[pos0][k].evaluate(v) == 0) ++k;
				if (k < n) pos1 = pos0;				
			}
			else
			{
				if (pos2 == ~0u)
				{
					unsigned int k = 0;
					while (k < n && my_maps[pos0][k].evaluate(v) == 0) ++k;
					if (k < n)
					{
						k = 0;
						while (k < n && (my_maps[pos0][k].evaluate(v) ^ my_maps[pos1][k].evaluate(v)) == 0) ++k;
						if (k < n) pos2 = pos0;
					}
				}
				else
				{
					unsigned int k = 0;
					while (k < n && my_maps[pos0][k].evaluate(v) == 0) ++k;
					if (k < n)
					{
						k = 0;
						while (k < n && (my_maps[pos0][k].evaluate(v) ^ my_maps[pos1][k].evaluate(v)) == 0) ++k;
						if (k < n)
						{
							k = 0;
							while (k < n && (my_maps[pos0][k].evaluate(v) ^ my_maps[pos2][k].evaluate(v)) == 0) ++k;
							if (k < n)
							{
								k = 0;
								while (k < n && (my_maps[pos0][k].evaluate(v) ^ my_maps[pos1][k].evaluate(v) ^ my_maps[pos2][k].evaluate(v)) == 0) ++k;
								if (k < n) return list<vector<uint8_t>> ();
							}
						}
					}
				}
			}
		}
		return res;		
	}
	else
	{
		list<vector<uint8_t>> res;
		if (pos1 == ~0u)
		{
			{
				auto sys = my_system;
				for (unsigned int k = 0; k < n; k++) sys.emplace_back(my_maps[pos0][k]);
				auto tmp = FindFirstLayerSTEP1(my_maps,n,pos0+1,pos1,pos2,sys);
				res.splice(res.begin(),tmp);
			}
			{
				auto tmp = FindFirstLayerSTEP1(my_maps,n,pos0+1,pos0,pos2,my_system);
				res.splice(res.begin(),tmp);
			}
		}
		else
		{
			if (pos2 == ~0u)
			{
				for (unsigned int i = 0; i <= 1; i++)
				{
					auto sys = my_system;
					for (unsigned int k = 0; k < n; k++)
					{
						Poly p = my_maps[pos0][k];
						if (i == 1) p += my_maps[pos1][k];
						sys.emplace_back(p);
					}
					auto tmp = FindFirstLayerSTEP1(my_maps,n,pos0+1,pos1,pos2,sys);
					res.splice(res.begin(),tmp);				
				}
				{
					auto tmp = FindFirstLayerSTEP1(my_maps,n,pos0+1,pos1,pos0,my_system);
					res.splice(res.begin(),tmp);
				}
			}
			else
			{
				for (unsigned int i = 0; i <= 1; i++)
				{
					for (unsigned int j = 0; j <= 1; j++)
					{
						auto sys = my_system;
						for (unsigned int k = 0; k < n; k++)
						{
							Poly p = my_maps[pos0][k];
							if (i == 1) p += my_maps[pos1][k];
							if (j == 1) p += my_maps[pos2][k];
							sys.emplace_back(p);
						}
						auto tmp = FindFirstLayerSTEP1(my_maps,n,pos0+1,pos1,pos2,sys);
						res.splice(res.begin(),tmp);				
					}
				}
			}
		}
		return res;
	}
}

list<vector<Poly>> FindFirstLayer(list<Poly> L, const unsigned int n)
{
	vector<map<unsigned int, Poly>> my_maps;
	for (auto const & P : L)
	{
		map<unsigned int, Poly> my_map;
		auto const & T = P.getTerms();
		for (auto const & t : T)
		{
			if (t.degree() == 2)
			{
				auto const & V = t.getVars();
				auto itV = V.begin();
				const unsigned int v1 = *itV;
				++itV;
				const unsigned int v2 = *itV;
				my_map[v1] += Term(v2);
				my_map[v2] += Term(v1);
			}
		}
		my_maps.emplace_back(move(my_map));
	}
	
	auto const nb_poly = my_maps.size();
	

	uniform_int_distribution<unsigned int> distribution(0,nb_poly-1);
	
	list<Poly> tmp;
	auto S = FindFirstLayerSTEP1(my_maps,n,0,~0u,~0u,tmp);
	cout << endl << "S contains " << S.size() << " vectors (should be close to " << n << ")" << endl;
	
	//unsigned int cpt_S = 0;
	vector<vector<uint8_t>> vecS;
	for (auto & v : S)
	{
		//cout << " " << cpt_S++ << ") ";
		//for (auto const & x : v) cout << static_cast<unsigned int>(x);
		//cout << endl;
		vecS.emplace_back(move(v));
	}
	//cout << endl;
	auto const size_vecS = vecS.size();
	
	map<vector<uint8_t>,list<pair<unsigned int, unsigned int>>> map_SS;
	list<vector<uint8_t>> SS;
	{
		for (unsigned int i1 = 0; i1 < size_vecS; ++i1)
		{
			auto const & v1 = vecS[i1];
			for (unsigned int i2 = 0; i2 < i1; ++i2)
			{
				vector<uint8_t> tmp;
				tmp.reserve(n);
				auto const & v2 = vecS[i2];
				for (unsigned int i = 0; i < n; ++i) tmp.emplace_back(v1[i] ^ v2[i]);
				map_SS[tmp].emplace_back(make_pair(i1,i2));
				SS.emplace_back(move(tmp));
			}
		}
	}
	
	cout << "SS contains " << SS.size() << " vectors " << flush;
		
	for (unsigned int k = 0; k < n; ++k)
	{
		list<vector<uint8_t>> SS_tmp;
		unsigned int my_bits[4];
		{
			set<unsigned int> bits;
			while (bits.size() != 4) bits.emplace(distribution(rand_generator));
			auto it = bits.begin();
			for (unsigned int i = 0; i < 4; i++)
			{
				my_bits[i] = *it;
				++it;
			}
		}
		unsigned int cst[5];
		for (cst[0] = 0; cst[0] <= 1; ++cst[0])
		{
			for (cst[1] = 0; cst[1] <= 1; ++cst[1])
			{
				for (cst[2] = 0; cst[2] <= 1; ++cst[2])
				{
					for (cst[3] = 0; cst[3] <= 1; ++cst[3])
					{
						{
							if (cst[0] == 0 && cst[1] == 0 && cst[2] == 0 && cst[3] == 0) continue;
							list<Poly> my_system;
							for (unsigned int i = 0; i < n; ++i)
							{
								Poly p;
								for (unsigned int j = 0; j < 4; ++j)
								{
									if (cst[j] != 0) p += my_maps[my_bits[j]][i];
								}
								my_system.emplace_back(move(p));
							}
							auto const end = my_system.end();
							auto itSS = SS.begin();
							auto const endSS = SS.end();
							while (itSS != endSS)
							{
								auto it = my_system.begin();
								while (it != end && it->evaluate(*itSS) == 0) ++it;
								if (it == end)
								{
									SS_tmp.emplace_back(move(*itSS));
									itSS = SS.erase(itSS);
								}
								else ++itSS;
							}
						}
					}
				}
			}
		}
		SS = move(SS_tmp);
	}
	
	cout << " -> only " << SS.size() << " possible (should be close to 2x" << n << ")" << endl;
	
	map<unsigned int, set<unsigned int>> map_cycles;
	map<unsigned int, unsigned int> map_valid;
	
	for (auto const & v : SS)
	{
		auto const & L = map_SS[v];
		for (auto const & my_pair : L)
		{
			map_cycles[my_pair.first].emplace(my_pair.second);
			map_cycles[my_pair.second].emplace(my_pair.first);
		}
	}
	map_SS.clear();
	list<list<unsigned int>> my_cycles;
	for (unsigned int i = 0; i < size_vecS; i++) map_valid[i] = 0;
	
	
	list<vector<Poly>> res;
	unsigned int cpt_order = 0;
	for (unsigned int i = 0; i <= size_vecS - n; i++)
	{
		if (map_valid[i] == 0)
		{
			auto const & s = map_cycles[i];
			map_valid[i] = 1;
			for (auto const x : s)
			{
				if (map_valid[x] == 0)
				{
					map_valid[x] = 2;
					auto tmp = FindFirstLayerCHAINS(x,i,3,n,map_cycles,map_valid);
					for (auto & l : tmp)
					{
						bool mat_ok = true;
						l.emplace_back(i);
						cout << " - order " << ++cpt_order << ": ";
						for (auto const y : l) cout << y << " ";
						cout << endl << endl;
						Matrix<uint8_t> mat1 (n,n,0);
						Matrix<uint8_t> mat2 (n,n,0);
						for (unsigned int i = 0; i < n; ++i) mat1(i,i) = 1;
						auto it = l.begin();
						for (unsigned int i = 0; i < n; ++i)
						{
							auto const & v = vecS[*it];
							for (unsigned int j = 0; j < n; ++j) mat2(j,i) = v[j];
							++it;
						}
						cout << "corresponding matrix: ";
						for (unsigned int j = 0; j < n; ++j)
						{
							unsigned int i = j;
							while (i < n && mat2(i,j) == 0) ++i;
							if (i == n)
							{
								mat_ok = false;
								break;
							}
							mat2.swapLines(i,j);
							mat1.swapLines(i,j);
							for (i = 0; i < n; ++i)
							{
								if (mat2(i,j) != 0 && i != j)
								{
									for (unsigned int k = 0; k < n; k++)
									{
										mat2(i,k) ^= mat2(j,k);
										mat1(i,k) ^= mat1(j,k);
									}
								}
							}
						}
						if (mat_ok == true)
						{
							cout << endl;
							for (unsigned int i = 0; i < n; ++i)
							{
								for (unsigned int j = 0; j < n; ++j) cout << static_cast<unsigned int>(mat1(i,j));
								cout << endl;
							}
							cout << endl;
							vector<Poly> L;
							L.reserve(n);
							for (unsigned int i = 0; i < n; ++i)
							{
								Poly p;
								for (unsigned int j = 0; j < n; ++j)
								{
									if (mat1(i,j) == 1) p += Term(j);
								}
								L.emplace_back(move(p));
							}
							res.emplace_back(move(L));
						}
						else
						{
							cout << "non valid" << endl;
						}
						
						
						my_cycles.emplace_back(move(l));
					}
					map_valid[x] = 0;
				}
			}			
		}
		map_valid[i] = ~0;
	}
	
	return res;
}

void FindSecondLayer(vector<Poly> L1, list<Poly> L2, const unsigned int n)
{
	vector<map<Term, Poly>> my_maps;
	my_maps.reserve(n);
	for (unsigned int i = 0; i < n; ++i)
	{
		map<Term, Poly> my_map;
		auto P = L1[i] + (L1[(i+1)%n] * (L1[(i+2)%n] + 1));
		auto const & terms = P.getTerms();
		for (auto const & t : terms) my_map[t] = Term();
		my_map[Term()] += Term(i) + Term((i+1)%n) + Term((i+1)%n)*Term((i+2)%n);
		auto const & terms1 = L1[(i+1)%n].getTerms();
		for (auto const & t : terms1) my_map[t] += Term((i+2)%n);
		auto const & terms2 = L1[(i+2)%n].getTerms();
		for (auto const & t : terms2) my_map[t] += Term((i+1)%n);
		my_maps.emplace_back(move(my_map));
	}
	
	vector<map<Term, Poly>> my_eqs;
	for (auto const & p : L2)
	{
		
	}
	
}



int main(int argc, char **argv)
{
	unsigned int n = 127;
	unsigned int nb_perturbations = 115;
	
	if (PRINT_INFO) cout << "Generating public key " << flush;
	vector<Poly> public_key = GeneratePublicKey(n,nb_perturbations); // return ASA
	if (PRINT_INFO) cout << ": done. " << endl;
	
	list<Poly> L;
	for (unsigned int i = 0; i < (n + 2) - nb_perturbations; ++i) L.emplace_back(move(public_key[i])); // keep only the n - p + 2 first polynomials
	
	FindFirstLayer(L,n); // retreive the first linear layer
	
	cout << "done";
	
	return 0;
}

