#include <stdio.h>
#include <stdlib.h>

#define SIZE 127

// -- end param --

#define BLOCK __uint128_t
#define ull   unsigned long long
#define uchar unsigned char

#define SWAP(a,b) (a) ^= (b); (b) ^= (a); (a) ^= (b);

#define RAND (((BLOCK) (random()&0xffff))<<(14*8) | ((BLOCK) (random()&0xffff))<<(12*8) \
  | ((BLOCK) (random()&0xffff))<<(10*8) | ((BLOCK) (random()&0xffff))<<(8*8) \
  | ((BLOCK) (random()&0xffff))<<(6*8)  | ((BLOCK) (random()&0xffff))<<(4*8) \
  | ((BLOCK) (random()&0xffff))<<(2*8)  | ((BLOCK) (random()&0xffff))<<(0*8))

void print128(BLOCK x) {
  printf("%016llx %016llx",(ull) (x>>64),(ull) x);
}

void sort2(int *i, int *j) {
  if (*i < *j) {
    SWAP(*i,*j);
  }
}

void sort4(int *i, int *j, int *k, int *l) {
  sort2(i,j);
  sort2(j,k);
  sort2(k,l);
  sort2(i,j);
  sort2(j,k);
  sort2(i,j);
}

uchar hw(BLOCK x) {
  int i;
  uchar count = 0;
  for (i=0;i<SIZE;i++) {
    count += x&1;
    x >>= 1;
  }
  return count;
}

// -------------------------------------------------- quadratic poly type

typedef struct {
  BLOCK data[SIZE];
} quad_t;

uchar quad_equal(quad_t *a, quad_t *b) {
  int i;
  for (i=0;i<SIZE;i++) {
    if (a->data[i] != b->data[i])
      return 0;
  }
  return 1;
}

void quad_zero(quad_t *q) {
  int i;
  for (i=0;i<SIZE;i++)
    q->data[i] = 0;
}

void quad_copy(quad_t *r, quad_t *q) {
  int i;
  for (i=0;i<SIZE;i++)
    r->data[i] = q->data[i];
}

uchar quad_get(quad_t *q, int i, int j) {
  sort2(&i,&j);
  return (q->data[i]>>j)&1;
}


void quad_set(quad_t *q, int i, int j) {
  sort2(&i,&j);
  q->data[i] |= ((BLOCK) 1)<<j;
}

void quad_flip(quad_t *q, int i, int j) {
  sort2(&i,&j);
  q->data[i] ^= ((BLOCK) 1)<<j;
}

void quad_symmetrize(quad_t *q) {
  BLOCK r;
  int i,k;
  for (k=0;k<SIZE;k++) {
    r = 0;
    for (i=0;i<SIZE;i++) {
      if (quad_get(q,k,i))
	r |= ((BLOCK) 1)<<i;
    }
    q->data[k] = r & ~(((BLOCK) 1)<<k);
  }
}

void quad_random(quad_t *q) {
  int i,j;
  for (i=0;i<SIZE;i++)
    q->data[i] = RAND;
  quad_symmetrize(q);
}

void quad_print(quad_t *q) {
  int i,j;
  uchar f = 0;
  for (i=0;i<SIZE;i++)
    for (j=0;j<i;j++)
      if (quad_get(q,i,j)) {
	if (f)
	  printf(" + ");
	printf("x%02Xx%02X",i,j);
	f = 1;
      }
}

// -------------------------------------------------- quartic poly type

typedef struct {
  BLOCK data[SIZE][SIZE][SIZE];
} quartic_t;

void quartic_new(quartic_t **q) {
  if (!(*q = calloc(sizeof(quartic_t),1))) {
    fprintf(stderr,"%s: malloc fail for quartic\n",__func__);
    exit(1);
  }
}

void quartic_zero(quartic_t *q) {
  int i,j,k;
  for (i=0;i<SIZE;i++)
    for (j=0;j<SIZE;j++)
      for (k=0;k<SIZE;k++)
	q->data[i][j][k] = 0;
}

void quartic_copy(quartic_t *r, quartic_t *q) {
  int i,j,k;
  for (i=0;i<SIZE;i++)
    for (j=0;j<SIZE;j++)
      for (k=0;k<SIZE;k++)
	r->data[i][j][k] = q->data[i][j][k];
}

uchar quartic_get(quartic_t *q, int i, int j, int k, int l) {
  sort4(&i,&j,&k,&l);
  return (q->data[i][j][k]>>l)&1;
}


void quartic_set(quartic_t *q, int i, int j, int k, int l) {
  sort4(&i,&j,&k,&l);
  q->data[i][j][k] |= ((BLOCK) 1)<<l;
}

void quartic_flip(quartic_t *q, int i, int j, int k, int l) {
  sort4(&i,&j,&k,&l);
  q->data[i][j][k] ^= ((BLOCK) 1)<<l;
}

void quartic_print(quartic_t *q) {
  int i,j,k,l;
  uchar f = 0;
  for (i=0;i<SIZE;i++)
    for (j=0;j<i;j++)
      for (k=0;k<j;k++)
	for (l=0;l<k;l++)
	  if (quartic_get(q,i,j,k,l)) {
	    if (f)
	      printf(" + ");
	    printf("x%02Xx%02Xx%02Xx%02X",i,j,k,l);
	    f = 1;
	  }
}

void qua_mult(quartic_t *r, quad_t *a, quad_t *b) {
  int i,j,k,l;
  for (i=0;i<SIZE;i++)
    for (j=0;j<i;j++)
      for (k=0;k<SIZE;k++)
	for (l=0;l<k;l++)
	  if (!(i==k) && !(i==l) && !(j==k) && !(j==l))
	    if (quad_get(a,i,j) && quad_get(b,k,l))
	      quartic_flip(r,i,j,k,l);
}

// -------------------------------------------------- temp

void quad_raw(quad_t *q) {
  int i;
  for (i=0;i<SIZE;i++) {
    printf("%2d: ",i);
    print128(q->data[i]);
    puts("");
  }
}

void test() {
  quartic_t q[1];
  quad_t a[1],b[1];
  quad_zero(a);
  quad_zero(b);
  quartic_zero(q);

  quad_set(a,11,5);
  quad_flip(a,12,11);

  quad_set(b,1,2);
  quad_set(b,15,5);

  qua_mult(q,a,b);

  quad_print(a);
  puts("");
  puts("*");
  quad_print(b);
  puts("");
  puts("=");
  quartic_print(q);
  puts("");
}

// -------------------------------------------------- attack components

uchar zero_oracle(quartic_t *q, int a, int b, int c) {
  int i;
  for (i=0;i<SIZE;i++) {
    if (i==a || i==b || i==c)
      continue;
    if (quartic_get(q,a,b,c,i))
      return 0;
  }
  return 1;
}

uchar zero_pair_oracle(quartic_t *q, int a, int b) {
  int i;
  for (i=0;i<SIZE;i++) {
    if (i==a || i==b)
      continue;
    if (zero_oracle(q,a,b,i))
      return 1;
  }
  return 0;
}

void find_zero(int *x, int *y, int index, quartic_t *q) {
  int a,b;
  for (;;) {
    a = (RAND % (SIZE-1))+1;
    b = RAND % a;
    if ((a==index) || (b==index))
      continue;
    if (zero_oracle(q,a,b,index)) {
      *x = a;
      *y = b;
      return;
    }
  }
}

void find_good(int *x, int *y, int index, quartic_t *q) {
  int a,c;
  for (;;) {
    a = (RAND % (SIZE-1))+1;
    if (!zero_pair_oracle(q,a,index))
      continue;
    c = RAND % a;
    if (!zero_pair_oracle(q,c,index) || zero_pair_oracle(q,a,c))
      continue;
    *x = a;
    *y = c;
    return;
  }
}

void get_space(BLOCK *ra, BLOCK *rb, int index, quartic_t *q) {
  int x,y;
  int i,j;
  BLOCK a = 0,b,c,r;

  for (;;) {
    find_good(&x,&y,index,q);
    r = 0;
    for (i=0;i<SIZE;i++)
      if (quartic_get(q,x,y,index,i))
	r |= ((BLOCK) 1)<<i;
    if (a && a^r) {
      *ra = a;
      *rb = r;
      return;
    }
    a = r;
  }
}

// -------------------------------------------------- full attack

uchar check_coherence(quad_t *M, int index) {
  int i;
  for (i=0;i<index;i++) {
    if (((M->data[index]>>i)&1) != ((M->data[i]>>index)&1))
      return 0;
  }
  return 1;
}

uchar full_attack_rec(quad_t *A, quad_t *B, int index, quartic_t *q) {
  BLOCK x, y;
  int try;
  if (index==SIZE)
    return 1;
  for (try=0;try<2;try++) {
    get_space(&x,&y,index,q);
    A->data[index] = x;
    B->data[index] = y;
    if (check_coherence(A,index) && check_coherence(B,index))
      if (full_attack_rec(A,B,index+1,q))
	return 1;
    A->data[index] = x;
    B->data[index] = x^y;
    if (check_coherence(A,index) && check_coherence(B,index))
      if (full_attack_rec(A,B,index+1,q))
	return 1;
    A->data[index] = y;
    B->data[index] = x;
    if (check_coherence(A,index) && check_coherence(B,index))
      if (full_attack_rec(A,B,index+1,q))
	return 1;
    A->data[index] = y;
    B->data[index] = x^y;
    if (check_coherence(A,index) && check_coherence(B,index))
      if (full_attack_rec(A,B,index+1,q))
	return 1;
    A->data[index] = x^y;
    B->data[index] = y;
    if (check_coherence(A,index) && check_coherence(B,index))
      if (full_attack_rec(A,B,index+1,q))
	return 1;
    A->data[index] = x^y;
    B->data[index] = x;
    if (check_coherence(A,index) && check_coherence(B,index))
      if (full_attack_rec(A,B,index+1,q))
	return 1;
  }
  return 0;
}

void full_attack(quad_t *ra, quad_t *rb, quartic_t *q) {
  if (full_attack_rec(ra,rb,0,q))
    puts("success");
  else
    puts("FAIL");
}

// -------------------------------------------------- test functions

void test_find_zero() {
  quad_t a[1];
  quad_t b[1];
  quartic_t *q;

  quartic_new(&q);
  quad_random(a);
  quad_random(b);
  qua_mult(q,a,b);
  
  int x,y,z = 0;
  find_zero(&x,&y,z,q);
  printf("%02x %02x %02x\n",x,y,z);
  puts("verification:");
  printf("a: %d %d %d\n",quad_get(a,x,y),quad_get(a,y,z),quad_get(a,x,z));
  printf("b: %d %d %d\n",quad_get(b,x,y),quad_get(b,y,z),quad_get(b,x,z));

  free(q);
}

void test_find_good() {
  quad_t a[1];
  quad_t b[1];
  quartic_t *q;

  quartic_new(&q);
  quad_random(a);
  quad_random(b);
  qua_mult(q,a,b);
  
  int x,y,z = 0;
  find_good(&x,&y,z,q);
  printf("%02x %02x %02x\n",x,y,z);
  puts("verification:");
  printf("a: %d %d %d\n",quad_get(a,x,y),quad_get(a,y,z),quad_get(a,x,z));
  printf("b: %d %d %d\n",quad_get(b,x,y),quad_get(b,y,z),quad_get(b,x,z));

  free(q);
}

void test_get_space() {
  quad_t a[1];
  quad_t b[1];
  quartic_t *q;
  quartic_new(&q);

  quad_random(a);
  quad_random(b);
  qua_mult(q,a,b);

  BLOCK ra, rb;
  int index = 64;

  get_space(&ra,&rb,index,q);
  print128(ra);
  puts("");
  print128(rb);
  puts("");
  puts("a=");
  print128(a->data[index]);
  puts("");
  puts("b=");
  print128(b->data[index]);
  puts("");
}

// -------------------------------------------------- main

int main(int argc, char **argv) {
  quad_t a[1];
  quad_t b[1];
  quartic_t *q;
  quartic_new(&q);

  unsigned seed = 1;
  if (argc > 1)
    seed = atoi(argv[1]);
  srandom(seed);
  printf("seed = %u\n",seed);

  quad_random(a);
  quad_random(b);

  qua_mult(q,a,b);

  quad_t ra[1], rb[1];

  full_attack(ra,rb,q);

  if (quad_equal(a,ra) || quad_equal(b,ra) || quad_equal(a,rb) || quad_equal(b,rb))
    puts("SUCCESS CONFIRMED");
  
  return 0;
}
